@@ -46,17 +46,35 @@ ensure_python_initialized <- function(required_module = NULL) {
 
     .globals$py_config <- initialize_python(required_module, use_environment)
 
-    # generate 'R' helper object
-    py_inject_r(envir = globalenv())
-
     # remap output streams to R output handlers
     remap_output_streams()
+    
+    # generate 'R' helper object
+    py_inject_r(envir = globalenv())
+    
+    # inject hooks
+    py_inject_hooks()
+    
+    # install required packages
+    configure_environment()
 
   }
 }
 
 initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
+  # disallow initialization of Python within .onLoad()
+  
+  # NOTE: disabled for this release of reticulate as there are too many packages
+  # already forcing initialization of Python on load
+  # check_forbidden_initialization()
+  
+  # provide hint to install Miniconda if no Python is found
+  python_not_found <- function(msg) {
+    hint <- "Use reticulate::install_miniconda() if you'd like to install a Miniconda Python environment."
+    stop(paste(msg, hint, sep = "\n"), call. = FALSE)
+  }
+
   # resolve top level module for search
   if (!is.null(required_module))
     required_module <- strsplit(required_module, ".", fixed = TRUE)[[1]][[1]]
@@ -66,12 +84,13 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
   # check for basic python prerequsities
   if (is.null(config)) {
-    stop("Installation of Python not found, Python bindings not loaded.")
+    python_not_found("Installation of Python not found, Python bindings not loaded.")
   } else if (!is_windows() && is.null(config$libpython)) {
-    stop("Python shared library not found, Python bindings not loaded.")
+    python_not_found("Python shared library not found, Python bindings not loaded.")
   } else if (is_incompatible_arch(config)) {
-    stop("Your current architecture is ", current_python_arch(), " however this version of ",
-         "Python is compiled for ", config$architecture, ".")
+    fmt <- "Your current architecture is %s; however, this version of Python was compiled for %s."
+    msg <- sprintf(fmt, current_python_arch(), config$architecture)
+    python_not_found(msg)
   }
 
   # check numpy version and provide a load error message if we don't satisfy it
@@ -85,7 +104,7 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
   # set as appropriate)
   if (nzchar(config$virtualenv))
     Sys.setenv(VIRTUAL_ENV = config$virtualenv)
-
+  
   # set R_SESSION_INITIALIZED flag (used by rpy2)
   curr_session_env <- Sys.getenv("R_SESSION_INITIALIZED", unset = NA)
   Sys.setenv(R_SESSION_INITIALIZED = sprintf('PID=%s:NAME="reticulate"', Sys.getpid()))
@@ -134,7 +153,9 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
       return()
 
     # if we loaded one of the requested versions, everything is ok
-    if (config$python %in% requested_versions)
+    actual <- normalizePath(config$python, winslash = "/", mustWork = FALSE)
+    requested <- normalizePath(requested_versions, winslash = "/", mustWork = FALSE)
+    if (actual %in% requested)
       return()
 
     # otherwise, warn that we were unable to honor their request
@@ -155,10 +176,61 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
     }
 
   })
+  
+  # notify front-end (if any) that Python has been initialized
+  callback <- getOption("reticulate.initialized")
+  if (is.function(callback))
+    callback()
 
   # return config
   config
 }
 
-
-
+check_forbidden_initialization <- function() {
+  
+  if (is_python_initialized())
+    return(FALSE)
+  
+  override <- getOption(
+    "reticulate.allow.package.initialization",
+    default = FALSE
+  )
+  
+  if (identical(override, TRUE))
+    return(FALSE)
+  
+  calls <- sys.calls()
+  frames <- sys.frames()
+  
+  for (i in seq_along(calls)) {
+  
+    call <- calls[[i]]
+    frame <- frames[[i]]
+    if (!identical(call[[1]], as.name("runHook")))
+      next
+    
+    bad <-
+      identical(call[[2]], ".onLoad") ||
+      identical(call[[2]], ".onAttach")
+    
+    if (!bad)
+      next
+    
+    pkgname <- tryCatch(
+        get("pkgname", envir = frame),
+        error = function(e) "<unknown>"
+      )
+      
+    fmt <- paste(
+      "package '%s' attempted to initialize Python in %s().",
+      "Packages should not initialize Python themselves; rather, Python should",
+      "be loaded on-demand as requested by the user of the package. Please see",
+      "vignette(\"python_dependencies\", package = \"reticulate\") for more details."
+    )
+    
+    msg <- sprintf(fmt, pkgname, call[[2]])
+    warning(msg)
+    
+  }
+  
+}
