@@ -97,7 +97,7 @@ new_stack <- function() {
 
 }
 
-py_compile_eval <- function(code) {
+py_compile_eval <- function(code, compile_mode="single") {
 
   builtins <- import_builtins(convert = TRUE)
   sys <- import("sys", convert = TRUE)
@@ -116,9 +116,9 @@ py_compile_eval <- function(code) {
   # newline follows the code to be submitted
   code <- sub("\\s*$", "\n", code)
 
-  # compile and eval the code -- using 'single' here ensures that Python
-  # auto-prints statements as they are evaluated
-  compiled <- builtins$compile(code, '<string>', 'single')
+  # compile and eval the code -- use 'single' to auto-print statements
+  # as they are evaluated, or 'exec' to avoid auto-print
+  compiled <- builtins$compile(code, '<string>', compile_mode)
   output <- py_capture_output(builtins$eval(compiled, globals, locals))
 
   # save the value that was produced
@@ -175,3 +175,53 @@ path_prepend <- function(entries) {
   oldpath
 }
 
+# note: normally, we'd like to compare paths with normalizePath() but
+# that does not normalize for case on Windows by default so we fall back
+# to a heuristic (note: false positives are possible but we can accept
+# those in the contexts where this function is used)
+file_same <- function(lhs, rhs) {
+  
+  # check if paths are identical as-is
+  if (identical(lhs, rhs))
+    return(TRUE)
+  
+  # check if paths are identical after normalization
+  lhs <- normalizePath(lhs, winslash = "/", mustWork = FALSE)
+  rhs <- normalizePath(rhs, winslash = "/", mustWork = FALSE)
+  if (identical(lhs, rhs))
+    return(TRUE)
+  
+  # check if file info is the same
+  lhsi <- c(file.info(lhs, extra_cols = FALSE))
+  rhsi <- c(file.info(rhs, extra_cols = FALSE))
+  fields <- c("size", "isdir", "mode", "mtime", "ctime")
+  if (identical(lhsi[fields], rhsi[fields]))
+    return(TRUE)
+  
+  # checks failed; return FALSE
+  FALSE
+  
+}
+
+# normalize a path without following symlinks
+canonical_path <- function(path) {
+  
+  # on windows we normalize the whole path to avoid
+  # short path components leaking in
+  if (is_windows()) {
+    normalizePath(path, winslash = "/", mustWork = FALSE)
+  } else {
+    file.path(
+      normalizePath(dirname(path), winslash = "/", mustWork = FALSE),
+      basename(path)
+    )
+  }
+  
+}
+
+enumerate <- function(x, f, ...) {
+  n <- names(x)
+  lapply(seq_along(x), function(i) {
+    f(n[[i]], x[[i]], ...)
+  })
+}
